@@ -4,6 +4,7 @@
 # Copyright (c) 2012-2020 Snowflake Computing Inc. All right reserved.
 #
 
+import copy
 import logging
 import os
 import re
@@ -15,7 +16,7 @@ from io import StringIO
 from logging import getLogger
 from threading import Lock
 from time import strptime
-from typing import Callable, Generator, Iterable
+from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union
 
 from . import errors, proxy
 from .auth import Auth
@@ -24,12 +25,14 @@ from .auth_idtoken import AuthByIdToken
 from .auth_keypair import AuthByKeyPair
 from .auth_oauth import AuthByOAuth
 from .auth_okta import AuthByOkta
+from .auth_usrpwdmfa import AuthByUsrPwdMfa
 from .auth_webbrowser import AuthByWebBrowser
 from .chunk_downloader import DEFAULT_CLIENT_PREFETCH_THREADS, MAX_CLIENT_PREFETCH_THREADS, SnowflakeChunkDownloader
-from .compat import IS_LINUX, IS_WINDOWS, PY_ISSUE_23517, urlencode
+from .compat import IS_LINUX, IS_WINDOWS, quote, urlencode
 from .constants import (
     PARAMETER_AUTOCOMMIT,
     PARAMETER_CLIENT_PREFETCH_THREADS,
+    PARAMETER_CLIENT_REQUEST_MFA_TOKEN,
     PARAMETER_CLIENT_SESSION_KEEP_ALIVE,
     PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY,
     PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL,
@@ -39,6 +42,7 @@ from .constants import (
     PARAMETER_SERVICE_NAME,
     PARAMETER_TIMEZONE,
     OCSPMode,
+    QueryStatus,
 )
 from .converter import SnowflakeConverter
 from .cursor import LOG_MAX_QUERY_LENGTH, SnowflakeCursor
@@ -61,6 +65,7 @@ from .network import (
     KEY_PAIR_AUTHENTICATOR,
     OAUTH_AUTHENTICATOR,
     REQUEST_ID,
+    USR_PWD_MFA_AUTHENTICATOR,
     ReauthenticationRequest,
     SnowflakeRestful,
 )
@@ -72,7 +77,7 @@ from .util_text import construct_hostname, parse_account, split_statements
 
 
 def DefaultConverterClass():
-    if PY_ISSUE_23517 or IS_WINDOWS:
+    if IS_WINDOWS:
         from .converter_issue23517 import SnowflakeConverterIssue23517
         return SnowflakeConverterIssue23517
     else:
@@ -140,6 +145,7 @@ DEFAULT_CONFIGURATION = {
     'disable_request_pooling': (False, bool),  # snowflake
     # enable temporary credential file for Linux, default false. Mac/Win will overlook this
     'client_store_temporary_credential': (False, bool),
+    'client_request_mfa_token': (False, bool),
     'use_openssl_only': (False, bool),  # only use openssl instead of python only crypto modules
 }
 
@@ -203,6 +209,9 @@ class SnowflakeConnection(object):
         self._errorhandler = Error.default_errorhandler
         self._lock_converter = Lock()
         self.messages = []
+        self._async_sfqids = set()
+        self._done_async_sfqids = set()
+        self.telemetry_enabled = False
         logger.info(
             "Snowflake Connector for Python Version: %s, "
             "Python Version: %s, Platform: %s",
@@ -218,10 +227,9 @@ class SnowflakeConnection(object):
         self.converter = None
         self.connect(**kwargs)
         self._telemetry = TelemetryClient(self._rest)
-        self.telemetry_enabled = False
         self.incident = IncidentAPI(self._rest)
 
-    def __del__(self):
+    def __del__(self):  # pragma: no cover
         try:
             self.close(retry=False)
         except Exception:
@@ -260,6 +268,7 @@ class SnowflakeConnection(object):
 
     @property
     def region(self):
+        warnings.warn("Region has been deprecated and will be removed in the near future", PendingDeprecationWarning)
         return self._region
 
     @property
@@ -428,7 +437,13 @@ class SnowflakeConnection(object):
             # close telemetry first, since it needs rest to send remaining data
             logger.info('closed')
             self._telemetry.close(send_on_close=retry)
-            self.rest.delete_session(retry=retry)
+            if self._all_async_queries_finished():
+                logger.info('No async queries seem to be running, deleting session')
+                self.rest.delete_session(retry=retry)
+            else:
+                logger.info('There are {} async queries still running, not deleting session'.format(
+                    len(self._async_sfqids)
+                ))
             self.rest.close()
             self._rest = None
             del self.messages[:]
@@ -465,8 +480,6 @@ class SnowflakeConnection(object):
             if e.sqlstate == SQLSTATE_FEATURE_NOT_SUPPORTED:
                 logger.debug("Autocommit feature is not enabled for this "
                              "connection. Ignored")
-            else:
-                raise e
 
     def commit(self):
         """Commits the current transaction."""
@@ -574,6 +587,8 @@ class SnowflakeConnection(object):
             auth_instance = AuthByKeyPair(self._private_key)
         elif self._authenticator == OAUTH_AUTHENTICATOR:
             auth_instance = AuthByOAuth(self._token)
+        elif self._authenticator == USR_PWD_MFA_AUTHENTICATOR:
+            auth_instance = AuthByUsrPwdMfa(self._password)
         else:
             # okta URL, e.g., https://<account>.okta.com/
             auth_instance = AuthByOkta(self.rest, self.application)
@@ -608,8 +623,13 @@ class SnowflakeConnection(object):
                 PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL] = \
                     self._client_store_temporary_credential if IS_LINUX else True
 
+        if self._authenticator == USR_PWD_MFA_AUTHENTICATOR:
+            self._session_parameters[
+                PARAMETER_CLIENT_REQUEST_MFA_TOKEN] = \
+                    self._client_request_mfa_token if IS_LINUX else True
+
         auth = Auth(self.rest)
-        auth.read_temporary_credential(self.host, self.account, self.user, self._session_parameters)
+        auth.read_temporary_credentials(self.host, self.user, self._session_parameters)
         self._authenticate(auth_instance)
 
         self._password = None  # ensure password won't persist
@@ -621,6 +641,9 @@ class SnowflakeConnection(object):
         if type(auth_instance) is AuthByWebBrowser:
             if self._rest.id_token is not None:
                 return AuthByIdToken(self._rest.id_token)
+        if type(auth_instance) is AuthByUsrPwdMfa:
+            if self._rest.mfa_token is not None:
+                auth_instance.set_mfa_token(self._rest.mfa_token)
         return auth_instance
 
     def __config(self, **kwargs):
@@ -666,7 +689,7 @@ class SnowflakeConnection(object):
         if self._numpy:
             try:
                 import numpy  # noqa: F401
-            except ModuleNotFoundError:
+            except ModuleNotFoundError:  # pragma: no cover
                 Error.errorhandler_wrapper(
                     self, None, ProgrammingError,
                     {
@@ -698,7 +721,8 @@ class SnowflakeConnection(object):
                 DEFAULT_AUTHENTICATOR,
                 EXTERNAL_BROWSER_AUTHENTICATOR,
                 KEY_PAIR_AUTHENTICATOR,
-                OAUTH_AUTHENTICATOR
+                OAUTH_AUTHENTICATOR,
+                USR_PWD_MFA_AUTHENTICATOR,
             ]:
                 self._authenticator = auth_tmp
 
@@ -769,10 +793,13 @@ class SnowflakeConnection(object):
             )
             self._use_openssl_only = (os.environ['SF_USE_OPENSSL_ONLY'] == 'True')
 
-    def cmd_query(self, sql, sequence_counter, request_id,
-                  binding_params=None,
+    def cmd_query(self,
+                  sql: str,
+                  sequence_counter: int,
+                  request_id: uuid.UUID,
+                  binding_params: Union[None, Tuple, Dict[str, Dict[str, str]]] = None,
                   is_file_transfer: bool = False,
-                  statement_params=None,
+                  statement_params: Optional[Dict[str, str]] = None,
                   is_internal: bool = False,
                   _no_results: bool = False,
                   _update_current_object: bool = True):
@@ -869,7 +896,9 @@ class SnowflakeConnection(object):
             session_parameters=self._session_parameters,
         )
 
-    def _process_params_qmarks(self, params, cursor=None):
+    def _process_params_qmarks(self,
+                               params: Union[List, Tuple, None],
+                               cursor: 'SnowflakeCursor' = None) -> Dict[str, Dict[str, str]]:
         if not params:
             return None
         processed_params = {}
@@ -883,7 +912,6 @@ class SnowflakeConnection(object):
             Error.errorhandler_wrapper(self, cursor,
                                        ProgrammingError,
                                        errorvalue)
-            return None
         for idx, v in enumerate(params):
             if isinstance(v, tuple):
                 if len(v) != 2:
@@ -897,7 +925,6 @@ class SnowflakeConnection(object):
                             'errno': ER_FAILED_PROCESSING_PYFORMAT,
                         }
                     )
-                    return None
                 processed_params[str(idx + 1)] = {
                     'type': v[0],
                     'value': self.converter.to_snowflake_bindings(
@@ -918,7 +945,6 @@ class SnowflakeConnection(object):
                             'errno': ER_NOT_IMPLICITY_SNOWFLAKE_DATATYPE
                         }
                     )
-                    return None
                 if isinstance(v, list):
                     vv = [
                         self.converter.to_snowflake_bindings(
@@ -934,7 +960,9 @@ class SnowflakeConnection(object):
                 logger.debug("idx: %s, type: %s", k, v.get('type'))
         return processed_params
 
-    def _process_params(self, params, cursor=None):
+    def _process_params(self,
+                        params: Union[Dict, List, Tuple, None],
+                        cursor: 'SnowflakeCursor' = None) -> Union[Tuple, Dict]:
         if params is None:
             return {}
         if isinstance(params, dict):
@@ -960,7 +988,10 @@ class SnowflakeConnection(object):
                                        ProgrammingError,
                                        errorvalue)
 
-    def __process_params_dict(self, params, cursor=None):
+    def __process_params_dict(self,
+                              params: Union[Dict, List, Tuple, None],
+                              cursor: 'SnowflakeCursor' = None) -> Dict:
+        # TODO this function could be reworked
         try:
             to_snowflake = self.converter.to_snowflake
             escape = self.converter.escape
@@ -976,7 +1007,7 @@ class SnowflakeConnection(object):
             return res
         except Exception as e:
             errorvalue = {
-                'msg': "Failed processing pyformat-parameters; {}".format(
+                'msg': "Failed processing pyformat-parameters: {}".format(
                     e),
                 'errno': ER_FAILED_PROCESSING_PYFORMAT}
             Error.errorhandler_wrapper(
@@ -1097,3 +1128,111 @@ class SnowflakeConnection(object):
         else:
             self.rollback()
         self.close()
+
+    def _get_query_status(self, sf_qid: str) -> Tuple[QueryStatus, Dict[str, Any]]:
+        """Retrieves the status of query with sf_qid and returns it with the raw response.
+
+        This is the underlying function used by the public get_status functions.
+
+        Args:
+            sf_qid: Snowflake query id of interest.
+
+        Raises:
+            ValueError: if sf_qid is not a valid UUID string.
+        """
+        try:
+            uuid.UUID(sf_qid)
+        except ValueError:
+            raise ValueError("Invalid UUID: '{}'".format(sf_qid))
+        logger.debug('get_query_status sf_qid=\'{}\''.format(sf_qid))
+
+        status = 'NO_DATA'
+        status_resp = self.rest.request('/monitoring/queries/' + quote(sf_qid),
+                                        method='get',
+                                        client='rest')
+        queries = status_resp['data']['queries']
+        if len(queries) > 0:
+            status = queries[0]['status']
+        status_ret = QueryStatus[status]
+        # If query was started by us and it has finished let's cache this info
+        if sf_qid in self._async_sfqids and not self.is_still_running(status_ret):
+            self._async_sfqids.remove(sf_qid)
+            self._done_async_sfqids.add(sf_qid)
+        return status_ret, status_resp
+
+    def get_query_status(self, sf_qid: str) -> QueryStatus:
+        """Retrieves the status of query with sf_qid.
+
+        Query status is returned as a QueryStatus.
+
+        Args:
+            sf_qid: Snowflake query id of interest.
+
+        Raises:
+            ValueError: if sf_qid is not a valid UUID string.
+        """
+        status, status_resp = self._get_query_status(sf_qid)
+        return status
+
+    def get_query_status_throw_if_error(self, sf_qid: str) -> QueryStatus:
+        """Retrieves the status of query with sf_qid as a QueryStatus and raises an exception if the query terminated with an error.
+
+        Query status is returned as a QueryStatus.
+
+        Args:
+            sf_qid: Snowflake query id of interest.
+
+        Raises:
+            ValueError: if sf_qid is not a valid UUID string.
+        """
+        status, status_resp = self._get_query_status(sf_qid)
+        queries = status_resp['data']['queries']
+        if self.is_an_error(status):
+            if sf_qid in self._async_sfqids:
+                self._async_sfqids.remove(sf_qid)
+            message = status_resp.get('message')
+            if message is None:
+                message = ''
+            code = status_resp.get('code')
+            if code is None:
+                code = -1
+            sql_state = None
+            if 'data' in status_resp:
+                message += queries[0].get('errorMessage', '') if len(queries) > 0 else ''
+                sql_state = status_resp['data'].get('sqlState')
+            Error.errorhandler_wrapper(self, None,
+                                       ProgrammingError,
+                                       {'msg': message,
+                                        'errno': int(code),
+                                        'sqlstate': sql_state,
+                                        'sfqid': sf_qid})
+        return status
+
+    @staticmethod
+    def is_still_running(status: QueryStatus) -> bool:
+        """Checks whether given status is currently running."""
+        return status in (
+                QueryStatus.RUNNING,
+                QueryStatus.RESUMING_WAREHOUSE,
+                QueryStatus.QUEUED,
+                QueryStatus.QUEUED_REPAIRING_WAREHOUSE,
+                QueryStatus.NO_DATA,
+        )
+
+    @staticmethod
+    def is_an_error(status: QueryStatus) -> bool:
+        """Checks whether given status means that there has been an error."""
+        return status in (
+            QueryStatus.ABORTING,
+            QueryStatus.FAILED_WITH_ERROR,
+            QueryStatus.ABORTED,
+            QueryStatus.FAILED_WITH_INCIDENT,
+            QueryStatus.DISCONNECTED,
+            QueryStatus.BLOCKED,
+        )
+
+    def _all_async_queries_finished(self) -> bool:
+        """Checks whether all async queries started by this Connection have finished executing."""
+        queries = copy.copy(self._async_sfqids)  # get_query_status might update _async_sfqids, let's copy the list
+        finished_async_queries = (not self.is_still_running(self.get_query_status(q)) for q in queries)
+        return all(finished_async_queries)
