@@ -10,11 +10,12 @@ from collections import namedtuple
 from logging import getLogger
 from typing import Any, Dict
 
-import requests
-
 from .compat import quote
 from .constants import HTTP_HEADER_CONTENT_ENCODING, SHA256_DIGEST, FileHeader, ResultStatus
 from .encryption_util import EncryptionMetadata
+from .vendored import requests
+
+logger = getLogger(__name__)
 
 GCS_METADATA_PREFIX = 'x-goog-meta-'
 GCS_METADATA_SFC_DIGEST = GCS_METADATA_PREFIX + 'sfc-digest'
@@ -50,7 +51,6 @@ class SnowflakeGCSUtil:
         Returns:
             The client to communicate with GCS.
         """
-        logger = getLogger(__name__)
         stage_credentials = stage_info['creds']
         security_token = stage_credentials.get('GCS_ACCESS_TOKEN')
 
@@ -82,8 +82,6 @@ class SnowflakeGCSUtil:
         Returns:
             None, if uploading was successful.
         """
-        logger = getLogger(__name__)
-
         upload_url = meta.get('presigned_url', None)
         access_token = None
 
@@ -204,8 +202,6 @@ class SnowflakeGCSUtil:
         Returns:
             None, if downloading was successful.
         """
-        logger = getLogger(__name__)
-
         download_url = meta.get('presigned_url', None)
         gcs_headers = None
         access_token = None
